from typing import Any

from . import __version__

# Set Ensembl REST URL
ensembl_default_url = "https://rest.ensembl.org"

# Ensembl API lookup table
# Specifies the functions relevant to the Ensembl REST server
ensembl_api_table: dict[str, Any] = {
    # Archive
    "getArchiveById": {
        "doc": "Uses the given identifier to return its latest version",
        "url": "/archive/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getArchiveByMultipleIds": {
        "doc": "Retrieve the latest version for a set of identifiers",
        "url": "/archive/id",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["id"],
    },
    # Comparative Genomics
    "getCafeGeneTreeById": {
        "doc": "Retrieves a cafe tree of the gene tree using the gene tree stable identifier",
        "url": "/cafe/genetree/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getCafeGeneTreeMemberBySymbol": {
        "doc": "Retrieves the cafe tree of the gene tree that contains the gene identified by a symbol",
        "url": "/cafe/genetree/member/symbol/{{species}}/{{symbol}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getCafeGeneTreeMemberById": {
        "doc": "Retrieves the cafe tree of the gene tree that contains the gene / transcript / translation stable identifier in the given species",
        "url": "/cafe/genetree/member/id/{{species}}/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getGeneTreeById": {
        "doc": "Retrieves a gene tree for a gene tree stable identifier",
        "url": "/genetree/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getGeneTreeMemberBySymbol": {
        "doc": "Retrieves the gene tree that contains the gene identified by a symbol",
        "url": "/genetree/member/symbol/{{species}}/{{symbol}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getGeneTreeMemberById": {
        "doc": "Retrieves the gene tree that contains the gene / transcript / translation stable identifier in the given species",
        "url": "/genetree/member/id/{{species}}/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getAlignmentByRegion": {
        "doc": "Retrieves genomic alignments as separate blocks based on a region and species",
        "url": "/alignment/region/{{species}}/{{region}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getHomologyById": {
        "doc": "Retrieves homology information (orthologs) by species and Ensembl gene id",
        "url": "/homology/id/{{species}}/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getHomologyBySymbol": {
        "doc": "Retrieves homology information (orthologs) by symbol",
        "url": "/homology/symbol/{{species}}/{{symbol}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # Cross References
    "getXrefsBySymbol": {
        "doc": """Looks up an external symbol and returns all Ensembl objects linked to it. """
        """This can be a display name for a gene/transcript/translation, """
        """a synonym or an externally linked reference. """
        """If a gene's transcript is linked to the supplied symbol """
        """the service will return both gene and transcript (it supports transient links).""",
        "url": "/xrefs/symbol/{{species}}/{{symbol}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getXrefsById": {
        "doc": "Perform lookups of Ensembl Identifiers and retrieve their external references in other databases",
        "url": "/xrefs/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getXrefsByName": {
        "doc": """Performs a lookup based upon the primary accession or display label of an external reference  """
        """and returning the information we hold about the entry""",
        "url": "/xrefs/name/{{species}}/{{name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # Information
    "getInfoAnalysis": {
        "doc": "List the names of analyses involved in generating Ensembl data.",
        "url": "/info/analysis/{{species}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoAssembly": {
        "doc": """List the currently available assemblies for a species, along with toplevel sequences, """
        """chromosomes and cytogenetic bands.""",
        "url": "/info/assembly/{{species}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoAssemblyRegion": {
        "doc": "Returns information about the specified toplevel sequence region for the given species.",
        "url": "/info/assembly/{{species}}/{{region_name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoBiotypes": {
        "doc": """List the functional classifications of gene models that Ensembl associates with a particular species. """
        """Useful for restricting the type of genes/transcripts retrieved by other endpoints.""",
        "url": "/info/biotypes/{{species}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoBiotypesByGroup": {
        "doc": """Without argument the list of available biotype groups is returned. With :group argument provided, list the properties of biotypes within that group. """
        """Object type (gene or transcript) can be provided for filtering.""",
        "url": "/info/biotypes/groups/{{group}}/{{object_type}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoBiotypesByName": {
        "doc": """List the properties of biotypes with a given name. """
        """Object type (gene or transcript) can be provided for filtering.""",
        "url": "/info/biotypes/name/{{name}}/{{object_type}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoComparaMethods": {
        "doc": "List all compara analyses available (an analysis defines the type of comparative data).",
        "url": "/info/compara/methods",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoComparaSpeciesSets": {
        "doc": "List all collections of species analysed with the specified compara method.",
        "url": "/info/compara/species_sets/{{methods}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoComparas": {
        "doc": """Lists all available comparative genomics databases and their data release. """
        """DEPRECATED: use info/genomes/division instead.""",
        "url": "/info/comparas",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoData": {
        "doc": """Shows the data releases available on this REST server. """
        """May return more than one release (unfrequent non-standard Ensembl configuration).""",
        "url": "/info/data",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoEgVersion": {
        "doc": "Returns the Ensembl Genomes version of the databases backing this service",
        "url": "/info/eg_version",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoExternalDbs": {
        "doc": "Lists all available external sources for a species.",
        "url": "/info/external_dbs/{{species}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoDivisions": {
        "doc": "Get list of all Ensembl divisions for which information is available",
        "url": "/info/divisions",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoGenomesByName": {
        "doc": "Find information about a given genome",
        "url": "/info/genomes/{{name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoGenomesByAccession": {
        "doc": "Find information about genomes containing a specified INSDC accession",
        "url": "/info/genomes/accession/{{accession}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoGenomesByAssembly": {
        "doc": "Find information about a genome with a specified assembly",
        "url": "/info/genomes/assembly/{{assembly_id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoGenomesByDivision": {
        "doc": "Find information about all genomes in a given division. May be large for Ensembl Bacteria.",
        "url": "/info/genomes/division/{{division}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoGenomesByTaxonomy": {
        "doc": "Find information about all genomes beneath a given node of the taxonomy",
        "url": "/info/genomes/taxonomy/{{taxon_name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoPing": {
        "doc": "Checks if the service is alive.",
        "url": "/info/ping",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoRest": {
        "doc": "Shows the current version of the Ensembl REST API.",
        "url": "/info/rest",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoSoftware": {
        "doc": "Shows the current version of the Ensembl API used by the REST server.",
        "url": "/info/software",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoSpecies": {
        "doc": "Lists all available species, their aliases, available adaptor groups and data release.",
        "url": "/info/species",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoVariationBySpecies": {
        "doc": "List the variation sources used in Ensembl for a species.",
        "url": "/info/variation/{{species}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoVariationConsequenceTypes": {
        "doc": "Lists all variant consequence types.",
        "url": "/info/variation/consequence_types",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoVariationPopulationIndividuals": {
        "doc": "List all individuals for a population from a species",
        "url": "/info/variation/populations/{{species}}/{{population_name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getInfoVariationPopulations": {
        "doc": "List all populations for a species",
        "url": "/info/variation/populations/{{species}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # Linkage Disequilibrium
    "getLdId": {
        "doc": """Computes and returns LD values between the given variant and all other variants """
        """in a window centered around the given variant. The window size is set to 500 kb.""",
        "url": "/ld/{{species}}/{{id}}/{{population_name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getLdPairwise": {
        "doc": "Computes and returns LD values between the given variants.",
        "url": "/ld/{{species}}/pairwise/{{id1}}/{{id2}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getLdRegion": {
        "doc": "Computes and returns LD values between all pairs of variants in the defined region.",
        "url": "/ld/{{species}}/region/{{region}}/{{population_name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # Lookup
    "getLookupById": {
        "doc": "Find the species and database for a single identifier e.g. gene, transcript, protein",
        "url": "/lookup/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getLookupByMultipleIds": {
        "doc": """Find the species and database for several identifiers. """
        """IDs that are not found are returned with no data.""",
        "url": "/lookup/id",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["ids"],
    },
    "getLookupBySymbol": {
        "doc": "Find the species and database for a symbol in a linked external database",
        "url": "/lookup/symbol/{{species}}/{{symbol}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getLookupByMultipleSymbols": {
        "doc": """Find the species and database for a set of symbols in a linked external database. """
        """Unknown symbols are omitted from the response.""",
        "url": "/lookup/symbol/{{species}}",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["symbols"],
    },
    # Mapping
    "getMapCdnaToRegion": {
        "doc": """Convert from cDNA coordinates to genomic coordinates. """
        """Output reflects forward orientation coordinates as returned from the Ensembl API.""",
        "url": "/map/cdna/{{id}}/{{region}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getMapCdsToRegion": {
        "doc": """Convert from CDS coordinates to genomic coordinates. """
        """Output reflects forward orientation coordinates as returned from the Ensembl API.""",
        "url": "/map/cds/{{id}}/{{region}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getMapAssemblyOneToTwo": {
        "doc": "Convert the co-ordinates of one assembly to another",
        "url": "/map/{{species}}/{{asm_one}}/{{region}}/{{asm_two}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getMapTranslationToRegion": {
        "doc": """Convert from protein (translation) coordinates to genomic coordinates. """
        """Output reflects forward orientation coordinates as returned from the Ensembl API.""",
        "url": "/map/translation/{{id}}/{{region}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # Ontologies and Taxonomy
    "getAncestorsById": {
        "doc": "Reconstruct the entire ancestry of a term from is_a and part_of relationships",
        "url": "/ontology/ancestors/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getAncestorsChartById": {
        "doc": "Reconstruct the entire ancestry of a term from is_a and part_of relationships.",
        "url": "/ontology/ancestors/chart/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getDescendantsById": {
        "doc": """Find all the terms descended from a given term. """
        """By default searches are conducted within the namespace of the given identifier""",
        "url": "/ontology/descendants/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getOntologyById": {
        "doc": "Search for an ontological term by its namespaced identifier",
        "url": "/ontology/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getOntologyByName": {
        "doc": "Search for a list of ontological terms by their name",
        "url": "/ontology/name/{{name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getTaxonomyClassificationById": {
        "doc": "Return the taxonomic classification of a taxon node",
        "url": "/taxonomy/classification/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getTaxonomyById": {
        "doc": "Search for a taxonomic term by its identifier or name",
        "url": "/taxonomy/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getTaxonomyByName": {
        "doc": "Search for a taxonomic id by a non-scientific name",
        "url": "/taxonomy/name/{{name}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # Overlap
    "getOverlapById": {
        "doc": """Retrieves features (e.g. genes, transcripts, variants and more) """
        """that overlap a region defined by the given identifier.""",
        "url": "/overlap/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getOverlapByRegion": {
        "doc": """Retrieves features (e.g. genes, transcripts, variants and more) """
        """that overlap a given region.""",
        "url": "/overlap/region/{{species}}/{{region}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getOverlapByTranslation": {
        "doc": """Retrieve features related to a specific Translation as described """
        """by its stable ID (e.g. domains, variants).""",
        "url": "/overlap/translation/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # Phenotype annotations
    "getPhenotypeByAccession": {
        "doc": "Return phenotype annotations for genomic features given a phenotype ontology accession",
        "url": "/phenotype/accession/{{species}}/{{accession}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getPhenotypeByGene": {
        "doc": "Return phenotype annotations for a given gene.",
        "url": "/phenotype/gene/{{species}}/{{gene}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getPhenotypeByRegion": {
        "doc": "Return phenotype annotations that overlap a given genomic region.",
        "url": "/phenotype/region/{{species}}/{{region}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getPhenotypeByTerm": {
        "doc": "Return phenotype annotations for genomic features given a phenotype ontology term",
        "url": "/phenotype/term/{{species}}/{{term}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # Regulation
    "getRegulationBindingMatrix": {
        "doc": "Return the specified binding matrix",
        "url": "/species/{{species}}/binding_matrix/{{binding_matrix}}/",
        "method": "GET",
        "content_type": "application/json",
    },
    # Sequences
    "getSequenceById": {
        "doc": """Request multiple types of sequence by stable identifier. """
        """Supports feature masking and expand options.""",
        "url": "/sequence/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getSequenceByMultipleIds": {
        "doc": "Request multiple types of sequence by a stable identifier list.",
        "url": "/sequence/id",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["ids"],
    },
    "getSequenceByRegion": {
        "doc": """Returns the genomic sequence of the specified region of the given species. """
        """Supports feature masking and expand options.""",
        "url": "/sequence/region/{{species}}/{{region}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getSequenceByMultipleRegions": {
        "doc": "Request multiple types of sequence by a list of regions.",
        "url": "/sequence/region/{{species}}",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["regions"],
    },
    # Transcript Haplotypes
    "getTranscriptHaplotypes": {
        "doc": "Computes observed transcript haplotype sequences based on phased genotype data",
        "url": "/transcript_haplotypes/{{species}}/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    # VEP
    "getVariantConsequencesByHGVSNotation": {
        "doc": "Fetch variant consequences based on a HGVS notation",
        "url": "/vep/{{species}}/hgvs/{{hgvs_notation}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getVariantConsequencesByMultipleHGVSNotations": {
        "doc": "Fetch variant consequences for multiple HGVS notations",
        "url": "/vep/{{species}}/hgvs/",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["hgvs_notations"],
    },
    "getVariantConsequencesById": {
        "doc": "Fetch variant consequences based on a variant identifier",
        "url": "/vep/{{species}}/id/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getVariantConsequencesByMultipleIds": {
        "doc": "Fetch variant consequences for multiple ids",
        "url": "/vep/{{species}}/id",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["ids"],
    },
    "getVariantConsequencesByRegion": {
        "doc": "Fetch variant consequences",
        "url": "/vep/{{species}}/region/{{region}}/{{allele}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getVariantConsequencesByMultipleRegions": {
        "doc": "Fetch variant consequences for multiple regions",
        "url": "/vep/{{species}}/region",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["variants"],
    },
    # Variation
    "getVariationRecoderById": {
        "doc": """Translate a variant identifier, HGVS notation or genomic SPDI notation """
        """to all possible variant IDs, HGVS and genomic SPDI""",
        "url": "/variant_recoder/{{species}}/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getVariationRecoderByMultipleIds": {
        "doc": """Translate a list of variant identifiers, HGVS notations or genomic SPDI """
        """notations to all possible variant IDs, HGVS and genomic SPDI""",
        "url": "/variant_recoder/{{species}}",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["ids"],
    },
    "getVariationById": {
        "doc": """Uses a variant identifier (e.g. rsID) to return the variation features """
        """including optional genotype, phenotype and population data""",
        "url": "/variation/{{species}}/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getVariationByPMCID": {
        "doc": """Uses a variant identifier (e.g. rsID) to return the variation features """
        """including optional genotype, phenotype and population data""",
        "url": "/variation/{{species}}/pmcid/{{pmcid}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getVariationByPMID": {
        "doc": """Uses a variant identifier (e.g. rsID) to return the variation features """
        """including optional genotype, phenotype and population data""",
        "url": "/variation/{{species}}/pmid/{{pmid}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getVariationByMultipleIds": {
        "doc": """Uses a list of variant identifiers (e.g. rsID) to return the variation features """
        """including optional genotype, phenotype and population data """,
        "url": "/variation/{{species}}",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["ids"],
    },
    # Variation GA4GH
    "getGA4GHBeacon": {
        "doc": "Return Beacon information",
        "url": "/ga4gh/beacon",
        "method": "GET",
        "content_type": "application/json",
    },
    "getGA4GHBeaconQuery": {
        "doc": "Return the Beacon response for allele information",
        "url": "/ga4gh/beacon/query?"
        "alternateBases={{alternateBases}};"
        "assemblyId={{assemblyId}};"
        "referenceBases={{referenceBases}};"
        "referenceName={{referenceName}};"
        "start={{start}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "postGA4GHBeaconQuery": {
        "doc": "Return the Beacon response for allele information",
        "url": "/ga4gh/beacon/query",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": [
            "alternateBases",
            "assemblyId",
            "end",
            "referenceBases",
            "referenceName",
            "start",
            "variantType",
        ],
    },
    "getGA4GHFeaturesById": {
        "doc": "Return the GA4GH record for a specific sequence feature given its identifier",
        "url": "/ga4gh/features/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "searchGA4GHFeatures": {
        "doc": "Return a list of sequence annotation features in GA4GH format",
        "url": "/ga4gh/features/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": [
            "end",
            "referenceName",
            "start",
            "featureSetId",
            "parentId",
        ],
    },
    "searchGA4GHCallset": {
        "doc": "Return a list of sets of genotype calls for specific samples in GA4GH format",
        "url": "/ga4gh/callsets/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["variantSetId", "name", "pageToken", "pageSize"],
    },
    "getGA4GHCallsetById": {
        "doc": "Return the GA4GH record for a specific CallSet given its identifier",
        "url": "/ga4gh/callsets/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "searchGA4GHDatasets": {
        "doc": "Return a list of datasets in GA4GH format",
        "url": "/ga4gh/datasets/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["pageToken", "pageSize"],
    },
    "getGA4GHDatasetsById": {
        "doc": "Return the GA4GH record for a specific dataset given its identifier",
        "url": "/ga4gh/datasets/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "searchGA4GHFeaturesets": {
        "doc": "Return a list of feature sets in GA4GH format",
        "url": "/ga4gh/featuresets/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["datasetId", "pageToken", "pageSize"],
    },
    "getGA4GHFeaturesetsById": {
        "doc": "Return the GA4GH record for a specific featureSet given its identifier",
        "url": "/ga4gh/featuresets/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "getGA4GHVariantsById": {
        "doc": "Return the GA4GH record for a specific variant given its identifier.",
        "url": "/ga4gh/variants/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "searchGA4GHVariantAnnotations": {
        "doc": "Return variant annotation information in GA4GH format for a region on a reference sequence",
        "url": "/ga4gh/variantannotations/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": [
            "variantAnnotationSetId",
            "effects",
            "end",
            "pageSize",
            "pageToken",
            "referenceId",
            "referenceName",
            "start",
        ],
    },
    "searchGA4GHVariants": {
        "doc": "Return variant call information in GA4GH format for a region on a reference sequence",
        "url": "/ga4gh/variants/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": [
            "variantSetId",
            "callSetIds",
            "referenceName",
            "start",
            "end",
            "pageToken",
            "pageSize",
        ],
    },
    "searchGA4GHVariantsets": {
        "doc": "Return a list of variant sets in GA4GH format",
        "url": "/ga4gh/variantsets/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["datasetId", "pageToken", "pageSize"],
    },
    "getGA4GHVariantsetsById": {
        "doc": "Return the GA4GH record for a specific VariantSet given its identifier",
        "url": "/ga4gh/variantsets/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "searchGA4GHReferences": {
        "doc": "Return a list of reference sequences in GA4GH format",
        "url": "/ga4gh/references/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": [
            "referenceSetId",
            "md5checksum",
            "accession",
            "pageToken",
            "pageSize",
        ],
    },
    "getGA4GHReferencesById": {
        "doc": "Return data for a specific reference in GA4GH format by id",
        "url": "/ga4gh/references/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "searchGA4GHReferencesets": {
        "doc": "Return a list of reference sets in GA4GH format",
        "url": "/ga4gh/referencesets/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["accession", "pageToken", "pageSize"],
    },
    "getGA4GHReferencesetsById": {
        "doc": "Return data for a specific reference set in GA4GH format",
        "url": "/ga4gh/referencesets/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
    "searchGA4GHVariantAnnotationsets": {
        "doc": "Return a list of annotation sets in GA4GH format",
        "url": "/ga4gh/variantannotationsets/search",
        "method": "POST",
        "content_type": "application/json",
        "post_parameters": ["variantSetId", "pageToken", "pageSize"],
    },
    "getGA4GHVariantAnnotationsetsById": {
        "doc": "Return meta data for a specific annotation set in GA4GH format",
        "url": "/ga4gh/variantannotationsets/{{id}}",
        "method": "GET",
        "content_type": "application/json",
    },
}

# Define HTTP status codes
ensembl_http_status_codes: dict[int, Any] = {
    200: (
        "OK",
        "Request was a success. Only process data from the service when you receive this code",
    ),
    400: (
        "Bad Request",
        "Occurs during exceptional circumstances such as the service is unable to find an ID. "
        "Check if the response Content-type or Accept was JSON. "
        "If so the JSON object is an exception hash with the message keyed under error",
    ),
    403: (
        "Forbidden",
        "You are submitting far too many requests and have been temporarily forbidden access to the service. "
        "Wait and retry with a maximum of 15 requests per second.",
    ),
    404: ("Not Found", "Indicates a badly formatted request. Check your URL"),
    408: ("Timeout", "The request was not processed in time. Wait and retry later"),
    415: (
        "Unsupported Media Type",
        "The server is refusing to service the request "
        "because the entity of the request is in a format not supported "
        "by the requested resource for the requested method",
    ),
    429: (
        "Too Many Requests",
        "You have been rate-limited; wait and retry. "
        "The headers X-RateLimit-Reset, X-RateLimit-Limit and X-RateLimit-Remaining will inform you "
        "of how long you have until your limit is reset and what that limit was. "
        "If you get this response and have not exceeded your limit then check "
        "if you have made too many requests per second.",
    ),
    500: (
        "Internal Server Error",
        "This error is not documented. Maybe there is an error in user input or "
        "the REST server could have problems. Try to do the query with curl. "
        "If your data input and query are correct, contact the Ensembl team",
    ),
    503: (
        "Service Unavailable",
        "The service is temporarily down; retry after a pause",
    ),
}

# Set the user agent
ensembl_user_agent = "pyEnsemblRest v" + __version__
ensembl_header = {"User-Agent": ensembl_user_agent}
ensembl_content_type = "application/json"

# Define known errors
ensembl_known_errors = [
    "something bad has happened",
    "Something went wrong while fetching from LDFeatureContainerAdaptor",
    "%s timeout" % ensembl_user_agent,
]



import importlib.metadata

__author__ = "Steve Moss"
__copyright__ = "Copyright (C) 2013-2024, Steve Moss"
__credits__ = ["Steve Moss"]
__license__ = "GNU GPLv3"
__version__: str = importlib.metadata.version("pyensemblrest")
__maintainer__ = "Steve Moss"
__email__ = "gawbul@gmail.com"
__status__ = "beta"

__all__ = [
    "EnsemblRest",
    "EnsemblRestError",
    "EnsemblRestRateLimitError",
    "EnsemblRestServiceUnavailable",
]

from .ensemblrest import EnsemblRest
from .exceptions import (
    EnsemblRestError,
    EnsemblRestRateLimitError,
    EnsemblRestServiceUnavailable,
)



from typing import Any

from requests.exceptions import ConnectionError

from .ensembl_config import ensembl_http_status_codes


class EnsemblRestError(Exception):
    """
    Generic error class, catch-all for most EnsemblRest issues.
    Special cases are handled by EnsemblRestRateLimitError and EnsemblRestServiceUnavailable.
    """

    def __init__(
        self,
        msg: str | ConnectionError,
        error_code: int | None = None,
        rate_reset: int | None = None,
        rate_limit: int | None = None,
        rate_remaining: int | None = None,
        retry_after: float | None = None,
    ) -> None:
        self.error_code = error_code

        if error_code is not None and error_code in ensembl_http_status_codes:
            msg = "EnsEMBL REST API returned a %s (%s): %s" % (
                error_code,
                ensembl_http_status_codes[error_code][0],
                msg,
            )

        super(EnsemblRestError, self).__init__(msg)

    @property
    def msg(self) -> Any:
        return self.args[0]


class EnsemblRestRateLimitError(EnsemblRestError):
    """
    Raised when you've hit a rate limit.
    The amount of seconds to retry your request in will be appended to the message.
    """

    def __init__(
        self,
        msg: str | ConnectionError,
        error_code: int | None = None,
        rate_reset: int | None = None,
        rate_limit: int | None = None,
        rate_remaining: int | None = None,
        retry_after: float | None = None,
    ) -> None:
        if isinstance(retry_after, float):
            msg = "%s (Rate limit hit:  Retry after %d seconds)" % (msg, retry_after)

        EnsemblRestError.__init__(self, msg, error_code=error_code)


class EnsemblRestServiceUnavailable(EnsemblRestError):
    """
    Raised when the service is down.
    """

    pass



import json
import logging
import re
import time
from typing import Any

import requests
from requests import Response
from requests.structures import CaseInsensitiveDict

# import ensemblrest modules
from .ensembl_config import (
    ensembl_api_table,
    ensembl_content_type,
    ensembl_default_url,
    ensembl_header,
    ensembl_http_status_codes,
    ensembl_known_errors,
    ensembl_user_agent,
)
from .exceptions import (
    EnsemblRestError,
    EnsemblRestRateLimitError,
    EnsemblRestServiceUnavailable,
)

# Logger instance
logger = logging.getLogger(__name__)


# FakeResponse object
class FakeResponse(object):
    def __init__(
        self,
        headers: CaseInsensitiveDict[str] | dict[str, Any],
        status_code: int,
        text: str,
    ):
        self.headers = headers
        self.status_code = status_code
        self.text: str = text


# EnsEMBL REST API object
class EnsemblRest(object):
    # class initialisation function
    def __init__(
        self, api_table: dict[str, Any] = ensembl_api_table, **kwargs: dict[str, Any]
    ) -> None:
        # read args variable into object as session_args
        self.session_args: dict[str, Any] = kwargs or {}

        # In order to rate limit the requests, like https://github.com/Ensembl/ensembl-rest/wiki/Example-Python-Client
        self.reqs_per_sec: int = 15
        self.req_count: int = 0
        self.last_req: float = 0
        self.wall_time: int = 1

        # get rate limit parameters, if provided
        self.rate_reset: int | None = None
        self.rate_limit: int | None = None
        self.rate_remaining: int | None = None
        self.rate_period: int | None = None
        self.retry_after: float | None = None

        # to record the last parameters used (in order to redo the query with an ensembl known error)
        self.last_url: str = ""
        self.last_headers: CaseInsensitiveDict[str] | dict[str, Any] = {}
        self.last_params: dict[str, Any] = {}
        self.last_data: dict[Any, Any] = {}
        self.last_method: str = ""
        self.last_attempt: int = 0
        self.last_response: Response | FakeResponse = Response()

        # the maximum number of attempts
        self.max_attempts: int = 5

        # setting a timeout
        self.timeout: int = 60

        # set default values if those values are not provided
        self.__set_default()

        # setup requests session
        self.session = requests.Session()

        # update headers
        self.__update_headers()

        # add class methods relying api_table
        self.__add_methods(api_table)

    def __set_default(self) -> None:
        """Set default values"""

        # initialise default values
        default_base_url = ensembl_default_url
        default_headers = ensembl_header
        default_content_type = ensembl_content_type
        default_proxies: dict[str, str] = {}

        if "base_url" not in self.session_args:
            self.session_args["base_url"] = default_base_url

        if "headers" not in self.session_args:
            self.session_args["headers"] = default_headers

        if "User-Agent" not in self.session_args["headers"]:
            self.session_args["headers"].update(default_headers)

        if "Content-Type" not in self.session_args["headers"]:
            self.session_args["headers"]["Content-Type"] = default_content_type

        if "proxies" not in self.session_args:
            self.session_args["proxies"] = default_proxies

    def __update_headers(self) -> None:
        """Update headers"""

        # update requests client with arguments
        client_args_copy = self.session_args.copy()
        for key, val in client_args_copy.items():
            if key in ("base_url", "proxies"):
                setattr(self.session, key, val)
                self.session_args.pop(key)

        # update headers as already exist within client
        self.session.headers.update(self.session_args.pop("headers"))

    def __add_methods(self, api_table: dict[str, Any]) -> None:
        """Add methods to class object"""

        # iterate over api_table keys and add key to class namespace
        for fun_name in api_table.keys():
            # setattr(self, key, self.register_api_func(key))
            # Not as a class attribute, but a class method
            self.__dict__[fun_name] = self.register_api_func(fun_name, api_table)

            # Set __doc__ for generic class method
            if "doc" in api_table[fun_name]:
                self.__dict__[fun_name].__doc__ = api_table[fun_name]["doc"]

            # add function name to the class methods
            self.__dict__[fun_name].__name__ = fun_name

    # dynamic api registration function
    def register_api_func(self, api_call: str, api_table: dict[str, Any]) -> Any:
        return lambda **kwargs: self.call_api_func(api_call, api_table, **kwargs)

    @staticmethod
    def __check_params(func: Any, kwargs: Any) -> list[Any]:
        """Check for mandatory parameters"""

        # Verify required variables and raise an Exception if needed
        mandatory_params = re.findall(r"\{\{(?P<m>[a-zA-Z1-9_]+)\}\}", func["url"])

        for param in mandatory_params:
            if param not in kwargs:
                logger.critical(
                    "'%s' param not specified. Mandatory params are %s"
                    % (param, mandatory_params)
                )
                raise Exception("mandatory param '%s' not specified" % param)
            else:
                logger.debug("Mandatory param %s found" % param)

        return mandatory_params

    # dynamic api call function
    def call_api_func(
        self, api_call: str, api_table: dict[str, Any], **kwargs: dict[str, Any]
    ) -> Any:
        # build url from api_table kwargs
        func = api_table[api_call]

        # check mandatory params
        mandatory_params = self.__check_params(func, kwargs)

        # resolving urls
        url = re.sub(
            r"\{\{(?P<m>[a-zA-Z1-9_]+)\}\}",
            lambda m: "%s" % kwargs.get(m.group(1)),
            self.session.base_url + func["url"],  # type: ignore[attr-defined]
        )

        # debug
        logger.debug("Resolved url: '%s'" % url)

        # Now I have to remove mandatory params from kwargs
        for param in mandatory_params:
            del kwargs[param]

        # Initialize with the ensembl default content type
        content_type: str | dict[str, Any] = ensembl_content_type

        # Override content type if it is defined by function
        if "content_type" in func:
            content_type = func["content_type"]

        # Ovveride content type if it is provied when calling function
        if "content_type" in kwargs:
            content_type = kwargs["content_type"]
            del kwargs["content_type"]

        # check the request type (GET or POST?)
        if func["method"] == "GET":
            logger.debug(
                "Submitting a GET request: url = '%s', headers = %s, params = %s"
                % (url, {"Content-Type": content_type}, kwargs)
            )

            # record this request
            self.last_url = url
            self.last_headers = {"Content-Type": content_type}
            self.last_params = kwargs
            self.last_data = {}
            self.last_method = "GET"
            self.last_attempt = 0

            resp = self.__get_response()

        elif func["method"] == "POST":
            # in a POST request, separate post parameters from other parameters
            data = {}

            # pass key=value in POST data from kwargs
            for key in func["post_parameters"]:
                if key in kwargs:
                    data[key] = kwargs[key]
                    del kwargs[key]

            logger.debug(
                "Submitting a POST request: url = '%s', headers = %s, params = %s, data = %s"
                % (url, {"Content-Type": content_type}, kwargs, data)
            )

            # record this request
            self.last_url = url
            self.last_headers = {"Content-Type": content_type}
            self.last_params = kwargs
            self.last_data = data
            self.last_method = "POST"
            self.last_attempt = 0

            resp = self.__get_response()

        else:
            raise NotImplementedError(
                "Method '%s' not yet implemented" % (func["method"])
            )

        # call response and return content
        return self.parseResponse(resp, content_type)

    # A function to get reponse from ensembl REST api
    def __get_response(self) -> Response | FakeResponse:
        """Call session get and post method. Return response"""

        # updating last_req time
        self.last_req = time.time()

        # Increment the request counter to rate limit requests
        self.req_count += 1

        # Evaluating the numer of request in a second (according to EnsEMBL rest specification)
        if self.req_count >= self.reqs_per_sec:
            delta = time.time() - self.last_req

            # sleep upto wall_time
            if delta < self.wall_time:
                to_sleep = self.wall_time - delta
                logger.debug("waiting %s" % to_sleep)
                time.sleep(to_sleep)

            self.req_count = 0

        # my response
        resp: Response | FakeResponse = Response()

        # deal with exceptions
        try:
            # another request using the correct method
            if self.last_method == "GET":
                resp = self.session.get(
                    self.last_url,
                    headers=self.last_headers,
                    params=self.last_params,
                    timeout=self.timeout,
                )
            elif self.last_method == "POST":
                # post parameters are load as POST data, other parameters are url parameters as GET requests
                resp = self.session.post(
                    self.last_url,
                    headers=self.last_headers,
                    data=json.dumps(self.last_data),
                    params=self.last_params,
                    timeout=self.timeout,
                )
            # other methods are verifiedby others functions

        except requests.ConnectionError as e:
            raise EnsemblRestServiceUnavailable(e)

        except requests.Timeout as e:
            logger.error("%s request timeout: %s" % (self.last_method, e))

            # create a fake response in order to redo the query
            resp = FakeResponse(
                headers=self.last_response.headers,
                status_code=400,
                text=json.dumps(
                    {"message": repr(e), "error": "%s timeout" % ensembl_user_agent}
                ),
            )

        # return response
        return resp

    # A function to deal with a generic response
    def parseResponse(
        self,
        resp: Response | FakeResponse,
        content_type: str | dict[str, Any] = "application/json",
    ) -> Any:
        """Deal with a generic REST response"""

        logger.debug("Got %s" % resp.text)

        # Record response for debug intent
        self.last_response = resp

        # Initialize some values. Check if I'm rate limited
        (
            self.rate_reset,
            self.rate_limit,
            self.rate_remaining,
            self.retry_after,
            self.rate_period,
        ) = self.__get_rate_limit(resp.headers)

        # parse status code
        if self.__check_retry(resp):
            return self.__retry_request()

        # Handle content in different way relying on content-type
        if content_type == "application/json":
            content = json.loads(resp.text)
        else:
            # Default
            content = resp.text

        return content

    def __check_retry(self, resp: Response | FakeResponse) -> bool:
        """Parse status code and print warnings. Return True if a retry is needed"""

        # default status code
        message = ensembl_http_status_codes[resp.status_code][1]

        # parse status codes
        if resp.status_code > 304:
            ExceptionType = EnsemblRestError

            # Try to derive a more useful message than ensembl default message
            if resp.status_code == 400:
                json_message = json.loads(resp.text)
                if "error" in json_message:
                    message = json_message["error"]

                # TODO: deal with special cases errors
                if message in ensembl_known_errors:
                    # call a function that will re-execute the REST request and then call again parseResponse
                    # if everithing is ok, a processed content is returned
                    logger.warning("EnsEMBL REST Service returned: %s" % message)

                    # return true if retry needed
                    return True
            elif resp.status_code == 500:
                # Retrying when we get a 500 error.
                # Due to Ensembl's condition on randomly returning 500s on valid requests.
                return True
            elif resp.status_code == 429:
                ExceptionType = EnsemblRestRateLimitError

            raise ExceptionType(
                message,
                error_code=resp.status_code,
                rate_reset=self.rate_reset,
                rate_limit=self.rate_limit,
                rate_remaining=self.rate_remaining,
                retry_after=self.retry_after,
            )

        # return a flag if status is ok
        return False

    @staticmethod
    def __get_rate_limit(
        headers: CaseInsensitiveDict[str] | dict[str, Any],
    ) -> tuple[int | None, int | None, int | None, float | None, int | None]:
        """Read rate limited attributes"""

        # initialize some values
        retry_after = None
        rate_reset = None
        rate_limit = None
        rate_remaining = None
        rate_period = None

        # for semplicity
        keys = [key.lower() for key in headers.keys()]

        if "X-RateLimit-Reset".lower() in keys:
            rate_reset = int(headers["X-RateLimit-Reset"])
            logger.debug("X-RateLimit-Reset: %s" % rate_reset)

        if "X-RateLimit-Period".lower() in keys:
            rate_period = int(headers["X-RateLimit-Period"])
            logger.debug("X-RateLimit-Period: %s" % rate_period)

        if "X-RateLimit-Limit".lower() in keys:
            rate_limit = int(headers["X-RateLimit-Limit"])
            logger.debug("X-RateLimit-Limit: %s" % rate_limit)

        if "X-RateLimit-Remaining".lower() in keys:
            rate_remaining = int(headers["X-RateLimit-Remaining"])
            logger.debug("X-RateLimit-Remaining: %s" % rate_remaining)

        if "Retry-After".lower() in keys:
            retry_after = float(headers["Retry-After"])
            logger.debug("Retry-After: %s" % retry_after)

        return rate_reset, rate_limit, rate_remaining, retry_after, rate_period

    def __retry_request(self) -> Any:
        """Retry last request in case of failure"""

        # update last attempt
        self.last_attempt += 1

        # a max of three attempts
        if self.last_attempt > self.max_attempts:
            # default status code
            message = ensembl_http_status_codes[self.last_response.status_code][1]

            # parse error if possible
            try:
                json_message = json.loads(self.last_response.text)
                if "error" in json_message:
                    message = json_message["error"]
            except ValueError:
                # In this case we didn't even get a JSON back.
                message = "Server returned invalid JSON."

            raise EnsemblRestError(
                "Max number of retries attempts reached. Last message was: %s"
                % message,
                error_code=self.last_response.status_code,
                rate_reset=self.rate_reset,
                rate_limit=self.rate_limit,
                rate_remaining=self.rate_remaining,
                retry_after=self.retry_after,
            )

        # sleep a while. Increment on each attempt
        to_sleep = (self.wall_time + 1) * self.last_attempt

        logger.debug("Sleeping %s" % to_sleep)
        time.sleep(to_sleep)

        # another request using the correct method
        if self.last_method == "GET":
            # debug
            logger.debug(
                "Retring last GET request (%s/%s): url = '%s', headers = %s, params = %s"
                % (
                    self.last_attempt,
                    self.max_attempts,
                    self.last_url,
                    self.last_headers,
                    self.last_params,
                )
            )

            resp = self.__get_response()

        elif self.last_method == "POST":
            # debug
            logger.debug(
                "Retring last POST request (%s/%s): url = '%s', headers = %s, params = %s, data = %s"
                % (
                    self.last_attempt,
                    self.max_attempts,
                    self.last_url,
                    self.last_headers,
                    self.last_params,
                    self.last_data,
                )
            )

            resp = self.__get_response()
        else:
            raise NotImplementedError(
                "Method '%s' not yet implemented" % (self.last_method)
            )

        # call response and return content
        return self.parseResponse(resp, self.last_headers["Content-Type"])

    def get_user_agent(self) -> str:
        """Return the pyEnsemblRest user agent"""
        return ensembl_user_agent
